# -*- coding: utf-8 -*-

# system imports
import sys
import os
import time
from traceback import format_exception
from subprocess import Popen
from datetime import timedelta, datetime
from shlex import quote

# external packages
import click
from PyQt6 import QtCore, QtWidgets, QtGui

# maestral modules
from maestral.constants import (
    IDLE,
    SYNCING,
    PAUSED,
    CONNECTING,
    CONNECTED,
    SYNC_ERROR,
    ERROR,
)
from maestral.daemon import (
    start_maestral_daemon_process,
    stop_maestral_daemon_process,
    MaestralProxy,
    Start,
    CommunicationError,
)
from maestral.exceptions import (
    KeyringAccessError,
    NoDropboxDirError,
    TokenExpiredError,
    TokenRevokedError,
    MaestralApiError,
    UpdateCheckError,
    NotLinkedError,
)

# local imports
from . import __url__
from .setup_dialog import SetupDialog
from .relink_dialog import RelinkDialog
from .dropbox_location_dialog import DropboxLocationDialog
from .settings_window import SettingsWindow
from .activity_window import ActivityWindow
from .sync_issues_window import SyncIssueWindow
from .resources import system_tray_icon, APP_ICON_PATH
from .utils import (
    BackgroundTask,
    MaestralBackgroundTask,
    elide_string,
    markup_urls,
    IS_MACOS,
    IS_BUNDLE,
)
from .widgets import (
    BackgroundTaskProgressDialog,
    UserDialog,
    show_dialog,
    show_stacktrace_dialog,
    show_update_dialog,
)
from .autostart import AutoStart


# noinspection PyTypeChecker,PyArgumentList
class MaestralGuiApp(QtWidgets.QSystemTrayIcon):
    """A Qt GUI for the Maestral daemon."""

    mdbx = None
    _started = False

    _context_menu_visible = False

    PAUSE_TEXT = "Pause Syncing"
    RESUME_TEXT = "Resume Syncing"

    icon_mapping = {
        IDLE: "idle",
        SYNCING: "syncing",
        PAUSED: "paused",
        CONNECTED: "idle",
        CONNECTING: "disconnected",
        SYNC_ERROR: "info",
        ERROR: "error",
    }

    def __init__(self, config_name="maestral"):
        super().__init__()

        self.config_name = config_name

        self._n_sync_errors = None
        self._current_icon = None

        self.settings_window = None
        self.sync_issues_window = None
        self.rebuild_dialog = None
        self._progress_dialog = None

        self.statusAction = None
        self.accountEmailAction = None
        self.accountUsageAction = None
        self.syncIssuesAction = None
        self.pauseAction = None
        self.activityAction = None

        self.loading_done = False

        self.autostart = AutoStart(self.config_name)

        self.icons = self.load_tray_icons()
        self.setIcon(CONNECTING)
        self.show_when_systray_available()

        self.menu = QtWidgets.QMenu()
        self.menu.setSeparatorsCollapsible(False)
        self.setContextMenu(self.menu)

        self.setup_ui_unlinked()

        self.check_for_updates_timer = QtCore.QTimer()
        self.check_for_updates_timer.timeout.connect(self.auto_check_for_updates)
        self.check_for_updates_timer.start(30 * 60 * 1000)  # every 30 min

        self.menu.aboutToShow.connect(self._onContextMenuAboutToShow)
        self.menu.aboutToHide.connect(self._onContextMenuAboutToHide)

        # schedule periodic updates
        self._wait_for_status = MaestralBackgroundTask(
            parent=self,
            config_name=self.config_name,
            target="status_change_longpoll",
            autostart=False,
        )
        self._wait_for_status.sig_result.connect(self.update_ui)

    def setIcon(self, icon_name):
        icon = self.icons.get(icon_name, self.icons[SYNCING])
        if self._current_icon != icon_name:
            self._current_icon = icon_name
            super().setIcon(icon)

    def _onContextMenuAboutToShow(self):
        self._context_menu_visible = True
        if self.loading_done:
            self.update_status()

    def _onContextMenuAboutToHide(self):
        self._context_menu_visible = False

    def update_ui(self):
        if self.loading_done:
            try:
                self.update_status()
                self.update_error()
            except CommunicationError:
                self.quit()
            else:
                self._wait_for_status.start()

    def show_when_systray_available(self):
        # If available, show icon, otherwise, set a timer to check back later.
        # This is a workaround for https://bugreports.qt.io/browse/QTBUG-61898
        if self.isSystemTrayAvailable():
            super().setIcon(self.icon())  # reload icon
            self.show()
        else:
            QtCore.QTimer.singleShot(1000, self.show_when_systray_available)

    def load_tray_icons(self):
        icons = dict()
        for key in self.icon_mapping:
            icons[key] = system_tray_icon(self.icon_mapping[key], self.geometry())

        return icons

    def load_maestral(self):
        self.mdbx = self.get_or_start_maestral_daemon()

        try:
            pending_link = self.mdbx.pending_link
        except KeyringAccessError:
            self.update_error()
            return

        pending_folder = self.mdbx.pending_dropbox_folder

        if pending_link or pending_folder:
            self.loading_done = SetupDialog.configureMaestral(self.mdbx)
        else:
            self.loading_done = True

        if self.loading_done:
            self.setup_ui_linked()
            self.mdbx.start_sync()
        else:
            self.quit()

    def get_or_start_maestral_daemon(self):
        res = start_maestral_daemon_process(self.config_name)

        if res == Start.Failed:
            title = "Could not start Maestral"
            message = (
                "Could not start or connect to sync daemon. Please try again "
                "and contact the developer if this issue persists."
            )
            show_dialog(title, message, level="error")
            self.quit()
        elif res == Start.AlreadyRunning:
            self._started = False
        elif res == Start.Ok:
            self._started = True

        return MaestralProxy(self.config_name)

    def setup_ui_unlinked(self):
        self.setToolTip("Not linked.")
        self.menu.clear()

        # ------------- populate context menu -------------------
        openDropboxFolderAction = self.menu.addAction("Open Dropbox Folder")
        openDropboxFolderAction.setEnabled(False)
        openWebsiteAction = self.menu.addAction("Launch Dropbox Website")
        openWebsiteAction.triggered.connect(self.on_website_clicked)

        self.menu.addSeparator()

        statusAction = self.menu.addAction("Setting up...")
        statusAction.setEnabled(False)

        self.menu.addSeparator()

        autostartAction = self.menu.addAction("Start on login")
        autostartAction.setCheckable(True)
        autostartAction.setChecked(self.autostart.enabled)
        autostartAction.triggered.connect(self.autostart.toggle)
        helpAction = self.menu.addAction("Help Center")
        helpAction.triggered.connect(self.on_help_clicked)

        self.menu.addSeparator()

        quitAction = self.menu.addAction("Quit Maestral")
        quitAction.triggered.connect(self.quit)

    def setup_ui_linked(self):
        self.autostart = None
        self.settings_window = SettingsWindow(self, self.mdbx)

        self.setToolTip(IDLE)

        # ------------- populate context menu -------------------

        self.menu.clear()

        openDropboxFolderAction = self.menu.addAction("Open Dropbox Folder")
        openDropboxFolderAction.triggered.connect(self.on_folder_clicked)
        openWebsiteAction = self.menu.addAction("Launch Dropbox Website")
        openWebsiteAction.triggered.connect(self.on_website_clicked)

        self.menu.addSeparator()

        self.accountEmailAction = self.menu.addAction(
            self.mdbx.get_state("account", "email")
        )
        self.accountEmailAction.setEnabled(False)

        self.accountUsageAction = self.menu.addAction(
            self.mdbx.get_state("account", "usage")
        )
        self.accountUsageAction.setEnabled(False)

        self.menu.addSeparator()

        self.statusAction = self.menu.addAction(IDLE)
        self.statusAction.setEnabled(False)
        self.pauseAction = self.menu.addAction(
            self.RESUME_TEXT if self.mdbx.paused else self.PAUSE_TEXT
        )
        self.pauseAction.triggered.connect(self.on_start_stop_clicked)

        self.activityAction = self.menu.addAction("Show Recent Changes...")
        self.activityAction.triggered.connect(self.on_activity_clicked)

        self.menu.aboutToShow.connect(self.update_snoozed)

        self.menu.addSeparator()

        self.snoozeMenu = self.menu.addMenu("Snooze Notifications")
        self.snooze30 = self.snoozeMenu.addAction("For the next 30 minutes")
        self.snooze60 = self.snoozeMenu.addAction("For the next hour")
        self.snooze480 = self.snoozeMenu.addAction("For the next 8 hours")

        def snooze_for(minutes):
            self.mdbx.notification_snooze = minutes

        self.snooze30.triggered.connect(lambda: snooze_for(30))
        self.snooze60.triggered.connect(lambda: snooze_for(60))
        self.snooze480.triggered.connect(lambda: snooze_for(480))
        self.snoozeSeparator = QtGui.QAction()
        self.snoozeSeparator.setSeparator(True)

        self.resumeNotificationsAction = QtGui.QAction("Turn on notifications")
        self.resumeNotificationsAction.triggered.connect(lambda: snooze_for(0))

        self.syncIssuesAction = self.menu.addAction("Show Sync Issues...")
        self.syncIssuesAction.triggered.connect(self.on_sync_issues_clicked)
        rebuildAction = self.menu.addAction("Rebuild index...")
        rebuildAction.triggered.connect(self.on_rebuild_clicked)

        self.menu.addSeparator()

        preferencesAction = self.menu.addAction("Preferences...")
        preferencesAction.triggered.connect(self.on_settings_clicked)
        updatesAction = self.menu.addAction("Check for Updates...")
        updatesAction.triggered.connect(self.on_check_for_updates_clicked)
        helpAction = self.menu.addAction("Help Center")
        helpAction.triggered.connect(self.on_help_clicked)

        self.menu.addSeparator()

        if self._started:
            quitAction = self.menu.addAction("Quit Maestral")
        else:
            quitAction = self.menu.addAction("Quit Maestral GUI")
        quitAction.triggered.connect(self.quit)

        # --------------- switch to idle icon -------------------
        self.setIcon(IDLE)

        # ------------ subscribe to status updates --------------
        self._wait_for_status.start()

    # callbacks for user interaction

    def auto_check_for_updates(self):
        last_update_check = self.mdbx.get_state("app", "update_notification_last")
        interval = self.mdbx.get_conf("app", "update_notification_interval")
        if interval == 0:  # checks disabled
            return
        elif time.time() - last_update_check > interval:
            checker = MaestralBackgroundTask(
                self, self.mdbx.config_name, "check_for_updates"
            )
            checker.sig_result.connect(self._notify_updates_auto)

    def on_check_for_updates_clicked(self):
        checker = MaestralBackgroundTask(
            self, self.mdbx.config_name, "check_for_updates"
        )
        self._progress_dialog = BackgroundTaskProgressDialog("Checking for Updates")
        self._progress_dialog.show()
        self._progress_dialog.rejected.connect(checker.sig_result.disconnect)

        checker.sig_result.connect(self._progress_dialog.accept)
        checker.sig_result.connect(self._notify_updates_user_requested)

    def _notify_updates_user_requested(self, res):
        if isinstance(res, UpdateCheckError):
            show_dialog(res.title, res.message, level="warning")
        elif res.update_available:
            show_update_dialog(res.latest_release, res.release_notes)
        else:
            message = "Maestral v{} is the newest version available.".format(
                res.latest_release
            )
            show_dialog("You’re up-to-date!", message)

    def _notify_updates_auto(self, res):
        if res.update_available:
            self.mdbx.set_state("app", "update_notification_last", time.time())
            show_update_dialog(res.latest_release, res.release_notes)

    def on_website_clicked(self):
        """Open the Dropbox website."""
        click.launch("https://www.dropbox.com/")

    def on_folder_clicked(self):
        """Open the Dropbox folder."""
        click.launch(self.mdbx.dropbox_path)

    def on_help_clicked(self):
        """Open the Dropbox help website."""
        click.launch(f"{__url__}/docs")

    def on_start_stop_clicked(self):
        """Pause / resume syncing on menu item clicked."""
        try:
            if self.pauseAction.text() == self.PAUSE_TEXT:
                self.mdbx.stop_sync()
                self.pauseAction.setText(self.RESUME_TEXT)
            elif self.pauseAction.text() == self.RESUME_TEXT:
                self.mdbx.start_sync()
                self.pauseAction.setText(self.PAUSE_TEXT)
            elif self.pauseAction.text() == "Start Syncing":
                self.mdbx.start_sync()
                self.pauseAction.setText(self.PAUSE_TEXT)
        except (NotLinkedError, NoDropboxDirError):
            self.restart()

    def on_settings_clicked(self):
        self.settings_window.show()
        self.settings_window.raise_()
        self.settings_window.activateWindow()

    def on_sync_issues_clicked(self):
        self.sync_issues_window = SyncIssueWindow(self.mdbx)
        self.sync_issues_window.show()
        self.sync_issues_window.raise_()
        self.sync_issues_window.activateWindow()
        self.sync_issues_window.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)

    def on_activity_clicked(self):
        self.activity_window = ActivityWindow(self.mdbx)
        self.activity_window.show()
        self.activity_window.raise_()
        self.activity_window.activateWindow()
        self.activity_window.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)

    def on_rebuild_clicked(self):
        self.rebuild_dialog = UserDialog(
            title="Rebuilt Maestral's sync index?",
            message=(
                "Rebuilding the index may take several minutes, depending on the size "
                "of your Dropbox. Any changes to local files will be synced once "
                "rebuilding has completed. If you quit Maestral during the process, "
                "rebuilding will  be resumed on the next launch."
            ),
            button_names=("Rebuild", "Cancel"),
        )
        res = self.rebuild_dialog.exec()
        if res == UserDialog.DialogCode.Accepted:
            self.mdbx.rebuild_index()

    # callbacks to update GUI

    def update_snoozed(self):
        minutes = self.mdbx.notification_snooze

        if minutes > 0:
            eta = datetime.now() + timedelta(minutes=minutes)

            self.snoozeMenu.setTitle(
                "Notifications snoozed until %s" % eta.strftime("%H:%M")
            )
            self.snoozeMenu.insertAction(self.snooze30, self.resumeNotificationsAction)
            self.snoozeMenu.insertAction(self.snooze30, self.snoozeSeparator)
        else:
            self.snoozeMenu.removeAction(self.resumeNotificationsAction)
            self.snoozeMenu.removeAction(self.snoozeSeparator)
            self.snoozeMenu.setTitle("Snooze Notifications")

    def update_status(self):
        """Change icon according to status."""
        n_sync_errors = len(self.mdbx.sync_errors)
        status = self.mdbx.status
        is_paused = self.mdbx.paused

        # update icon
        if n_sync_errors > 0 and status == IDLE:
            new_icon = SYNC_ERROR
        else:
            new_icon = status

        self.setIcon(new_icon)

        # update action texts
        if self.contextMenuVisible():
            if n_sync_errors > 0:
                self.syncIssuesAction.setText(
                    "Show Sync Issues ({0})...".format(n_sync_errors)
                )
            else:
                self.syncIssuesAction.setText("Show Sync Issues...")

            self.pauseAction.setText(self.RESUME_TEXT if is_paused else self.PAUSE_TEXT)
            self.accountUsageAction.setText(self.mdbx.get_state("account", "usage"))
            self.accountEmailAction.setText(self.mdbx.get_state("account", "email"))

            status_short = elide_string(status)
            self.statusAction.setText(status_short)

        # update tooltip
        self.setToolTip(status)

        # cache _n_errors
        self._n_sync_errors = n_sync_errors

    def update_error(self):
        errs = self.mdbx.fatal_errors

        if not errs:
            return
        else:
            self.mdbx.clear_fatal_errors()

        self.setIcon(ERROR)

        if self.pauseAction:
            self.pauseAction.setText(self.RESUME_TEXT)

        if self.statusAction:
            self.statusAction.setText(self.mdbx.status)

        err = errs[-1]

        if isinstance(err, NoDropboxDirError):
            # Show location dialog dialog.
            self._dbx_location_dialog = DropboxLocationDialog(self.mdbx)
            self._dbx_location_dialog.show()
            self._dbx_location_dialog.raise_()

        elif isinstance(err, (TokenRevokedError, TokenExpiredError)):
            # Show relink dialog.

            if isinstance(err, TokenExpiredError):
                reason = RelinkDialog.REVOKED
            else:
                reason = RelinkDialog.EXPIRED

            self._relink_dialog = RelinkDialog(self, reason)
            self._relink_dialog.show()
            self._relink_dialog.raise_()

        elif isinstance(err, MaestralApiError):
            # This is a known error. We show the error message and the corresponding
            # file path, if any.

            message = markup_urls(err.message)
            filename = err.dbx_path or err.local_path

            if filename:
                message = f"Path: {filename}\n{message}"

            show_dialog(err.title, message, level="error")

        else:
            # This is an unexpected error. We show the full stacktrace if available.
            if err.__traceback__:
                details = "".join(
                    format_exception(err.__class__, err, err.__traceback__)
                )
                show_stacktrace_dialog(details)
            else:
                show_dialog("An unexpected error occurred", str(err), level="error")

    def contextMenuVisible(self):
        return self._context_menu_visible

    def quit(self, *args, stop_daemon=None):
        """Quits Maestral.

        :param bool stop_daemon: If ``True``, the sync daemon will be PAUSED when
            quitting the GUI, if ``False``, it will be kept alive. If ``None``, the
            daemon will only be PAUSED if it was started by the GUI (default).
        """
        self._wait_for_status.cancel()

        # stop sync daemon if we started it or ``stop_daemon`` is ``True``
        if stop_daemon or self._started:
            task = BackgroundTask(
                parent=self,
                target=stop_maestral_daemon_process,
                args=(self.config_name,),
            )
            task.sig_result.connect(QtWidgets.QApplication.instance().quit)
        else:
            QtWidgets.QApplication.instance().quit()

    def restart(self):
        """Restarts the Maestral GUI and sync daemon."""
        pid = os.getpid()  # get ID of current process

        if IS_MACOS:
            restart_cmd = "lsof -p {0} +r 1 &>/dev/null; {1} -c {2}"
        else:
            restart_cmd = "tail --pid={0} -f /dev/null; {1} -c {2}"

        if IS_BUNDLE:
            launch_command = sys.executable
        else:
            launch_command = f"{sys.executable} -m maestral_qt"

        Popen(
            restart_cmd.format(pid, launch_command, quote(self.config_name)), shell=True
        )

        # quit Maestral
        self.quit(stop_daemon=True)


# noinspection PyArgumentList
def run(config_name="maestral"):
    """
    This is the main interactive entry point which starts the Qt GUI.

    :param str config_name: Name of Maestral config to run.
    """

    app = QtWidgets.QApplication(["Maestral"])
    app.setWindowIcon(QtGui.QIcon(APP_ICON_PATH))
    app.setQuitOnLastWindowClosed(False)

    maestral_gui = MaestralGuiApp(config_name)
    maestral_gui.load_maestral()
    sys.exit(app.exec())



# -*- coding: utf-8 -*-

# system imports
import sys
import os
import re
import platform
import traceback

# external packages
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtCore import Qt, QRect
from PyQt6.QtGui import QBrush, QImage, QPainter, QPixmap
from PyQt6.QtCore import pyqtSignal as Signal
from Pyro5.errors import ConnectionClosedError

# maestral modules
from maestral.daemon import MaestralProxy

# local imports
from .resources import rgb_to_luminance


THEME_DARK = "dark"
THEME_LIGHT = "light"

LINE_COLOR_DARK = (70, 70, 70)
LINE_COLOR_LIGHT = (213, 213, 213)

IS_BUNDLE = getattr(sys, "frozen", False)
IS_MACOS = platform.system() == "Darwin"
IS_LINUX = platform.system() == "Linux"
IS_MACOS_BUNDLE = IS_BUNDLE and IS_MACOS
IS_LINUX_BUNDLE = IS_BUNDLE and IS_LINUX


thread_pool = QtCore.QThreadPool()
thread_pool.setMaxThreadCount(10)


# ======================================================================================
# Helper functions
# ======================================================================================


def is_empty(dirname):
    """Checks if a directory is empty."""

    try:
        with os.scandir(dirname) as sciter:
            next(sciter)
    except StopIteration:
        return True

    return False


def elide_string(string, font=None, pixels=200, side="right"):
    """
    Elides a string to fit into the given width.

    :param str string: String to elide.
    :param font: Font to calculate size. If not given, the current style's default font
        for a QLabel is used.
    :param int pixels: Maximum width in pixels.
    :param str side: Side to truncate. Can be 'right' or 'left', defaults to 'right'.
    :return: Truncated string.
    :rtype: str
    """

    if not font:
        font = QtWidgets.QLabel().font()

    metrics = QtGui.QFontMetrics(font)
    mode = (
        Qt.TextElideMode.ElideRight if side == "right" else Qt.TextElideMode.ElideLeft
    )

    return metrics.elidedText(string, mode, pixels)


def get_scaled_font(scaling=1.0, bold=False, italic=False):
    """
    Returns the current style's default font for a QLabel but scaled by the given
    factor.

    :param float scaling: Scaling factor.
    :param bool bold: Sets the returned font to bold (defaults to ``False``)
    :param bool italic: Sets the returned font to italic (defaults to ``False``)
    :return: Scaled font.
    :rtype: QFont
    """
    label = QtWidgets.QLabel()
    font = label.font()
    font.setBold(bold)
    font.setItalic(italic)
    font_size = round(font.pointSize() * scaling)
    # noinspection PyTypeChecker
    font.setPointSize(font_size)

    return font


url_regex = re.compile(
    r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
)


def markup_urls(string):
    """
    Find any URLs in a string and wrap them in html a tags.

    :param str string: String to parse.
    :return: Marked-up string.
    """

    matches = url_regex.findall(string)

    for match in matches:
        url = "".join(match)
        href = f'<a href="{url}">{url}</a>'
        string = string.replace(url, href)

    return string


# noinspection PyArgumentList
def icon_to_pixmap(icon, width, height=None):
    """Converts a given icon to a pixmap. Automatically adjusts to high-DPI scaling.

    :param icon: Icon to convert.
    :param int width: Target point height.
    :param int height: Target point height.
    :return: ``QPixmap`` instance.
    """
    height = height or width
    return icon.pixmap(width, height)


# noinspection PyArgumentList
def center_window(widget):
    """
    Centers the given widget on screen.

    :param widget: QtWidgets.QWidget
    """
    screen = QtWidgets.QApplication.primaryScreen()
    geometry = screen.availableGeometry()

    x = (geometry.width() - widget.width()) // 2
    y = (geometry.height() - widget.height()) // 3

    widget.move(x, y)


# noinspection PyArgumentList, PyTypeChecker, PyCallByClass
def get_masked_image(path, size=64, overlay_text=""):
    """
    Returns a pixmap from an image file masked with a smooth circle.
    The returned pixmap will have a size of *size* × *size* pixels.

    :param str path: Path to image file.
    :param int size: Target size. Will be the diameter of the masked image.
    :param str overlay_text: Overlay text. This will be shown in white sans-serif on top
        of the image.
    :return: Masked image with overlay text.
    :rtype: QPixmap
    """

    with open(path, "rb") as f:
        imgdata = f.read()

    imgtype = path.split(".")[-1]

    # Load image and convert to 32-bit ARGB (adds an alpha channel):
    image = QImage.fromData(imgdata, imgtype)
    image.convertToFormat(QImage.Format.Format_ARGB32)

    # Crop image to a square:
    imgsize = min(image.width(), image.height())
    width = (image.width() - imgsize) / 2
    height = (image.height() - imgsize) / 2

    rect = QRect(
        round(width),
        round(height),
        imgsize,
        imgsize,
    )
    image = image.copy(rect)

    # Create the output image with the same dimensions and an alpha channel
    # and make it completely transparent:
    out_img = QImage(imgsize, imgsize, QImage.Format.Format_ARGB32)
    out_img.fill(Qt.GlobalColor.transparent)

    # Create a texture brush and paint a circle with the original image onto
    # the output image:
    brush = QBrush(image)  # Create texture brush
    painter = QPainter(out_img)  # Paint the output image
    painter.setBrush(brush)  # Use the image texture brush
    painter.setPen(Qt.PenStyle.NoPen)  # Don't draw an outline
    painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)  # Use AA
    painter.drawEllipse(0, 0, imgsize, imgsize)  # Actually draw the circle

    if overlay_text:
        # draw text
        font = QtGui.QFont("Arial Rounded MT Bold")
        font.setPointSize(imgsize * 0.4)
        painter.setFont(font)
        painter.setPen(Qt.GlobalColor.white)
        painter.drawText(
            QRect(0, 0, imgsize, imgsize), Qt.AlignmentFlag.AlignCenter, overlay_text
        )

    painter.end()  # We are done (segfault if you forget this)

    # Convert the image to a pixmap and rescale it.  Take pixel ratio into
    # account to get a sharp image on retina displays:
    pr = QtWidgets.QApplication.instance().devicePixelRatio()
    pm = QPixmap.fromImage(out_img)
    pm.setDevicePixelRatio(pr)
    size = int(pr * size)
    pm = pm.scaled(
        size,
        size,
        Qt.AspectRatioMode.KeepAspectRatio,
        Qt.TransformationMode.SmoothTransformation,
    )

    return pm


# noinspection PyArgumentList
def window_theme():
    """
    :return: THEME_LIGHT or THEME_DARK, corresponding to current user's UI theme.
    :rtype: bool
    """
    w = QtWidgets.QWidget()
    bg_color = w.palette().color(QtGui.QPalette.ColorRole.Window)
    bg_color_rgb = [bg_color.red(), bg_color.green(), bg_color.blue()]
    luminance = rgb_to_luminance(*bg_color_rgb)

    return THEME_LIGHT if luminance >= 0.4 else THEME_DARK


def is_dark_window():
    """Returns ``True`` if windows have a dark UI theme."""
    return window_theme() == THEME_DARK


# ======================================================================================
# Threading
# ======================================================================================


class WorkerEmitter(QtCore.QObject):
    sig_result = Signal(object)
    sig_done = Signal()


class Worker(QtCore.QRunnable):
    """A worker object. To be used in QThreads."""

    def __init__(self, target=None, args=None, kwargs=None):
        super().__init__()
        self._target = target
        self._args = args or ()
        self._kwargs = kwargs or {}
        self.emitter = WorkerEmitter()

    def run(self):
        try:
            res = self._target(*self._args, **self._kwargs)

            if hasattr(res, "__next__"):
                while True:
                    try:
                        next_res = next(res)
                        self.emitter.sig_result.emit(next_res)
                    except StopIteration:
                        return
            else:
                self.emitter.sig_result.emit(res)
        except Exception:
            traceback.print_exc()
        finally:
            self.emitter.sig_done.emit()


class MaestralWorker(Worker):
    """A worker object for Maestral. It uses a separate Maestral proxy to prevent
    the main connection from blocking."""

    def __init__(self, config_name="maestral", target=None, args=None, kwargs=None):
        self.config_name = config_name
        self.connection = None
        super().__init__(target, args, kwargs)

    def run(self):
        try:
            with MaestralProxy(self.config_name) as proxy:
                self.connection = proxy._m._pyroConnection

                func = proxy.__getattr__(self._target)
                res = func(*self._args, **self._kwargs)

                if hasattr(res, "__next__"):
                    while True:
                        try:
                            next_res = next(res)
                            self.emitter.sig_result.emit(next_res)
                        except StopIteration:
                            return
                else:
                    self.emitter.sig_result.emit(res)

        except ConnectionClosedError:
            pass
        except Exception as exc:
            print("".join(exc._pyroTraceback))
            print("{}: {}".format(type(exc).__name__, exc))
        finally:
            self.connection = None
            self.emitter.sig_done.emit()


class BackgroundTask(QtCore.QObject):
    """A utility class to manage a worker thread."""

    sig_result = Signal(object)
    sig_done = Signal()

    def __init__(
        self, parent=None, target=None, args=None, kwargs=None, autostart=True
    ):
        super().__init__(parent)
        self._target = target
        self._args = args or ()
        self._kwargs = kwargs or {}
        self.worker = None

        if autostart:
            self.start()

    def start(self):
        self.worker = Worker(target=self._target, args=self._args, kwargs=self._kwargs)
        self.worker.emitter.sig_result.connect(self.sig_result.emit)
        self.worker.emitter.sig_done.connect(self.sig_done.emit)
        thread_pool.start(self.worker)


class MaestralBackgroundTask(BackgroundTask):
    """A utility class to manage a worker thread. It uses a separate Maestral proxy
    to prevent the main connection from blocking."""

    def __init__(
        self,
        parent=None,
        config_name="maestral",
        target=None,
        args=None,
        kwargs=None,
        autostart=True,
    ):
        self.config_name = config_name
        super().__init__(parent, target, args, kwargs, autostart)

    def start(self):
        self.worker = MaestralWorker(
            config_name=self.config_name,
            target=self._target,
            args=self._args,
            kwargs=self._kwargs,
        )

        self.worker.emitter.sig_result.connect(self.sig_result.emit)
        self.worker.emitter.sig_done.connect(self.sig_done.emit)
        thread_pool.start(self.worker)

    def cancel(self):
        # brute force termination by closing the socket
        if self.worker and self.worker.connection:
            self.worker.connection.close()



# -*- coding: utf-8 -*-

# system imports
import argparse

# local imports
from maestral_qt.main import run


def main():
    """
    This is the main entry point for frozen executables.
    If only the --config-name option is given, it starts the GUI with the given config.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config-name", default="maestral")
    parsed_args = parser.parse_args()

    run(parsed_args.config_name)


if __name__ == "__main__":
    main()
